#include <SPI.h>
#include <ILI9341_game.h>
#include "pitches.h"

#define TFT_DC  2
#define TFT_CS 15
#define TFT_RST 4

ILI9341_game tft = ILI9341_game(TFT_CS, TFT_DC, TFT_RST);

int _width, _height;

int StickXPin = 34;
int StickYPin = 35;
int StickX;
int StickY;

int ButtonAPin = 32;
int ButtonBPin = 33;
int ButtonA;
int ButtonB;

int SoundPin = 25;
static int tone_init = 0;

#define SCREEN_WIDTH 240 // Display width, in pixels
#define SCREEN_HEIGHT 320 // Display height, in pixels

#define BLOCK  16
#define BLOCKX 10
#define BLOCKY 20
#define FRAMEX ((BLOCK * BLOCKX) + 2)
#define FRAMEY (BLOCK * BLOCKY)

int OldA;
int OldB;
int NewX;
int OldX;

int state;
int stage;
int score;
int count;
int tspeed;
int px, py, pt, pd;
int next_pt, next_pd;
int fall;
uint8_t block[BLOCKY][BLOCKX];

static const int tetrimino_data[7][4] = {
  0b0000000011110000,
  0b0010001000100010,
  0b0000000011110000,
  0b0010001000100010,

  0b0000011001100000,
  0b0000011001100000,
  0b0000011001100000,
  0b0000011001100000,

  0b0000000000110110,
  0b0000001000110001,
  0b0000000000110110,
  0b0000001000110001,

  0b0000000001100011,
  0b0000000100110010,
  0b0000000001100011,
  0b0000000100110010,

  0b0000010001110000,
  0b0000001100100010,
  0b0000000001110001,
  0b0000001000100110,

  0b0000000001110100,
  0b0000011000100010,
  0b0000000101110000,
  0b0000001000100011,

  0b0000001001110000,
  0b0000001000110010,
  0b0000000001110010,
  0b0000001001100010
 };

static const uint8_t tetrimino_color[7] = {
  5, // Cyan
  6, // Yellow
  4, // Greenyellow
  2, // Red
  1, // Blue
  7, // Orange
  3  // Purple
};
  
void setup() {

  tft.begin(40000000);
  tft.setRotation(0);
  tft.clearDisplay();
  tft.display();

  pinMode(ButtonAPin, INPUT_PULLUP);
  pinMode(ButtonBPin, INPUT_PULLUP);
  state=0;
}

void loop() {
  int x, y;
  int tet;
  int old_pd;
  int sc;
  StickX = analogRead(StickXPin)/4;
  StickY = analogRead(StickYPin)/4;
  ButtonA = digitalRead(ButtonAPin);
  ButtonB = digitalRead(ButtonBPin);

  sc=-1;
  if (state==0){
    show_title();
    if(ButtonA==0)
      state=1;
  }else if (state==1){
    count=0;
    px=3;
    py=0;
    pt=random(0, 7);
    pd=random(0, 4);
    next_pt=random(0, 7);
    next_pd=random(0, 4);
    stage=1;
    score=0;
    for(y=0;y<BLOCKY;y++){
      for(x=0;x<BLOCKX;x++){
        block[y][x]=0;
      }
    }
    state=2;
    tet=0;
    fall=0;
    tspeed=10;
  }else if (state==2){
    show_game();
    // Timer
    count++;
    if (score<20)
      tspeed=10;
    else if (score<40)
      tspeed=8;
    else if (score<60)
      tspeed=6;
    else
      tspeed=4;
    // Fall
    if (count>=tspeed || fall==1){
      count=0;
      sc=0;
      tet=fall_tet();
      if (tet==2)
        state=3;
      if (tet==1){
        px=3;
        py=0;
        pt=next_pt;
        pd=next_pd;
        next_pt=random(0, 7);
        next_pd=random(0, 4);
        fall=0;
        sc=check_block();
      }
    }
    // Move left/right
    if (StickX<256)
      NewX=-1;
    else if (StickX>768)
      NewX=1;
    else
      NewX=0;
    if (NewX<0 && OldX!=NewX){
//    if (NewX<0){
      if (move_tet(-1, 0)==0)
        px--;
    }
    if (NewX>0 && OldX!=NewX){
//    if (NewX>0){
      if (move_tet(1, 0)==0)
        px++;
    }
    // Round
    old_pd=pd;
    if (ButtonA==0 && OldA==1)
      pd++;
    if (ButtonB==0 && OldB==1)
      pd--;
    if (pd<0)
      pd=3;
    if (pd>3)
      pd=0;
    if (move_tet(0,0)!=0){
      if (move_tet(1,0)==0)
        px++;
      else if (move_tet(-1,0)==0)
        px--;
      else if (move_tet(2,0)==0)
        px+=2;
      else if (move_tet(-2,0)==0)
        px-=2;
      else
        pd=old_pd;
    }
    // Fall
    if (StickY<128)
      fall=1;
  // Game end
  }else if (state==3){
    show_end();
    if (ButtonA==0 && OldA==1)
      state=0;
    if (ButtonB==0 && OldB==1)
      state=0;
  }
  OldX=NewX;
  OldA=ButtonA;
  OldB=ButtonB;
  if (sc==0){
    tone(SoundPin, NOTE_C3, 10);
    delay(10);
    noTone(SoundPin);
    delay(40);
  }else if (sc==1){
    tone(SoundPin, NOTE_C5, 10);
    delay(10);
    noTone(SoundPin);
    delay(40);
  }else if (sc==2){
    tone(SoundPin, NOTE_C5, 10);
    delay(10);
    tone(SoundPin, NOTE_E5, 10);
    delay(10);
    noTone(SoundPin);
    delay(30);
  }else if (sc==3){
    tone(SoundPin, NOTE_C5, 10);
    delay(10);
    tone(SoundPin, NOTE_E5, 10);
    delay(10);
    tone(SoundPin, NOTE_G5, 10);
    delay(10);
    noTone(SoundPin);
    delay(20);
  }else if (sc==4){
    tone(SoundPin, NOTE_C5, 10);
    delay(10);
    tone(SoundPin, NOTE_E5, 10);
    delay(10);
    tone(SoundPin, NOTE_G5, 10);
    delay(10);
    tone(SoundPin, NOTE_C6, 10);
    delay(10);
    noTone(SoundPin);
    delay(10);
  }else{
    noTone(SoundPin);
//    delay(50);
  }
}

int move_tet(int dx, int dy)  // 0:Move  1:Stop  2:Error
{
  int i, j;
  int data;
  int sft;
  data=tetrimino_data[pt][pd];
  sft=data;
  for(j=0;j<4;j++){
    for(i=0;i<4;i++){
      if (sft & 1==1){
        // Error
        if (block[py+j][px+i]>0)
          return(2);
        // Stop
        if (block[py+j+dy][px+i+dx]>0 ||
            px+i+dx<0 || px+i+dx>=BLOCKX || py+j+dy>=BLOCKY){
          return(1);
        }
      }
      sft=sft >> 1;
    }
  }
  return(0);
}

int fall_tet()  // 0:Fall  1:Stop  2:Error
{
  int i, j;
  int data;
  int sft;
  data=tetrimino_data[pt][pd];
  sft=data;
  for(j=0;j<4;j++){
    for(i=0;i<4;i++){
      if (sft & 1==1){
        // Error
        if (block[py+j][px+i]>0)
          return(2);
        // Stop
        if (block[py+j+1][px+i]>0 || py+j+1>=BLOCKY){
          sft=data;
          for(j=0;j<4;j++){
            for(i=0;i<4;i++){
              if (sft & 1==1)
                block[py+j][px+i]=tetrimino_color[pt];
              sft=sft >> 1;
            }
          }
          return(1);
        }
      }
      sft=sft >> 1;
    }
  }
  py++;
  return(0);
}

int check_block()
{
  int i, x, y;
  int fill;
  int cy[4];
  int cc;
  for(x=0;x<4;x++)
    cy[x]=-1;
  cc=0;
  for(y=0;y<BLOCKY;y++){
    fill=1;
    for(x=0;x<BLOCKX;x++){
      if (block[y][x]==0){
        fill=0;
        break;
      }
    }
    if (fill==1){
      cy[cc]=y;
      cc++;
    }
  }
  // Clear block
  for(i=0;i<cc;i++){
    for(y=cy[i]-1;y>=0;y--){
      for(x=0;x<BLOCKX;x++)
        block[y+1][x]=block[y][x];
    }
  }
  // Score
  if (cc==1)
    score+=1;
  if (cc==2)
    score+=3;
  if (cc==3)
    score+=6;
  if (cc==4)
    score+=10;
  return(cc);
}

void show_title()
{
  tft.clearDisplay();
  tft.setTextSize(4);
  tft.setTextColor(ILI9341_WHITE);

  tft.setCursor(SCREEN_WIDTH/2-72, SCREEN_HEIGHT/4);     // Start at top-left corner
  tft.print(F("Tetris"));
  show_tetrimino(0, 10, tetrimino_data[0][1], tetrimino_color[0]);
  show_tetrimino(3, 10, tetrimino_data[1][1], tetrimino_color[1]);
  show_tetrimino(8, 10, tetrimino_data[2][1], tetrimino_color[2]);
  show_tetrimino(12, 10, tetrimino_data[3][1], tetrimino_color[3]);
  show_tetrimino(3, 15, tetrimino_data[4][1], tetrimino_color[4]);
  show_tetrimino(6, 15, tetrimino_data[5][1], tetrimino_color[5]);
  show_tetrimino(11, 15, tetrimino_data[6][1], tetrimino_color[6]);
  tft.display();
}

void show_end()
{
  tft.clearDisplay();
  tft.setTextSize(4);
  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(SCREEN_WIDTH/2-72, SCREEN_HEIGHT/4);
  tft.print(F("Tetris"));

  tft.setTextColor(ILI9341_GREEN);
  tft.setCursor(SCREEN_WIDTH/2-60, SCREEN_HEIGHT/2);
  tft.print(F("Score"));

  tft.setTextColor(ILI9341_WHITE);
  tft.setCursor(SCREEN_WIDTH/2-48, SCREEN_HEIGHT*3/4);
  tft.print(score, DEC);
  tft.display();
}

void show_game()
{
  int x, y;
  int cc;
  
  tft.clearDisplay();
  // Frame
  tft.drawLine(0, 0, 0, FRAMEY, ILI9341_WHITE);
  tft.drawLine(0, FRAMEY, FRAMEX, FRAMEY, ILI9341_WHITE);
  tft.drawLine(FRAMEX, 0, FRAMEX, FRAMEY, ILI9341_WHITE);
  // Tetrimino
  show_tetrimino(px, py, tetrimino_data[pt][pd], tetrimino_color[pt]);
  show_tetrimino(BLOCKX+1, BLOCKY-5, tetrimino_data[next_pt][next_pd], tetrimino_color[next_pt]);
  // Block
  for(y=0;y<BLOCKY;y++){
    for(x=0;x<BLOCKX;x++){
      if (block[y][x]>0){
        cc=t_color(block[y][x]);
        tft.fillRect(x*BLOCK+2, y*BLOCK, BLOCK-1, BLOCK-1, cc);
      }
    }
  }
  tft.setTextSize(2);
  tft.setCursor(FRAMEX+2, 4);     // Start at top-left corner
  tft.print(F("Tetris"));
  tft.setCursor(FRAMEX+2, 32);     // Start at top-left corner
  tft.print(score, DEC);
  // For debug
/*  
  tft.setCursor(FRAMEX+2, 40);     // Start at top-left corner
  tft.print(ballx, DEC);
  tft.setCursor(FRAMEX+2, 48);     // Start at top-left corner
  tft.print(bally, DEC);
  tft.setCursor(FRAMEX+2, 56);     // Start at top-left corner
  tft.print(balld, DEC);
*/
  tft.display();
}

void show_tetrimino(int tx, int ty, int data, uint8_t color)
{
  int i, j;
  int x, y;
  int sft;
  int cc;
  sft=data;
  for(j=0;j<4;j++){
    for(i=0;i<4;i++){
      if (sft & 1==1){
        x=tx+i;
        y=ty+j;
        cc=t_color(color);
        tft.fillRect(x*BLOCK+2, y*BLOCK, BLOCK-1, BLOCK-1, cc);
      }
      sft=sft >> 1;
    }
  }
}

int t_color(int color)
{
  switch(color){
    case 1 : return(ILI9341_BLUE);
    case 2 : return(ILI9341_RED);
    case 3 : return(ILI9341_MAGENTA);
    case 4 : return(ILI9341_GREENYELLOW);
    case 5 : return(ILI9341_CYAN);
    case 6 : return(ILI9341_YELLOW);
    default : return(ILI9341_ORANGE);
  }
}

void tone (uint8_t pin, uint16_t freq, uint16_t leng)
{
  if (tone_init==0){
    ledcSetup(1, 12000, 8);
    ledcAttachPin(pin, 1);
    tone_init=1;
  }
  ledcWriteTone(1, freq);
}

void noTone (uint8_t pin)
{
  ledcWriteTone(1, 1);
}
